@@ -4,7 +4,7 @@ import json
 import re
 import time
 from logging import error, info, warning
-from typing import Union, Any
+from typing import Any, Iterable, Union
 
 import act.api
 from act.api.re import UUID_MATCH
@@ -62,7 +62,7 @@ class RelevantFactBindings(ActBase):
     ]
 
     def __hash__(self):
-        return hash((self.__class__.__name__, self.name, self.fact_type))
+        return hash((self.__class__.__name__, self.name))
 
     def serialize(self):
         # Return None for empty objects (non initialized objects)
@@ -255,7 +255,7 @@ class AbstractFact(ActBase):
         Field(
             "type", deserializer=FactType, serializer=lambda fact_type: fact_type.name
         ),
-        Field("value", default=""),
+        Field("value", default=None),
         Field("id", serializer=False),
         Field("flags", serializer=False),
         Field("origin", deserializer=Origin),
@@ -501,6 +501,58 @@ class Fact(AbstractFact):
 
         return self
 
+    def format_objects(self, object_formatter=None):
+        if not object_formatter:
+            if not self.config or not self.config.object_formatter:
+                raise act.api.base.ArgumentError(
+                    "object_formatter must be specified as argument or in configuration"
+                )
+
+            object_formatter = self.config.object_formatter
+
+        if self.source_object:
+            self.source_object.value = object_formatter(
+                self.source_object.type.name, self.source_object.value
+            )
+        if self.destination_object:
+            self.destination_object.value = object_formatter(
+                self.destination_object.type.name,
+                self.destination_object.value,
+            )
+
+        return self
+
+    def validate_and_raise(self, object_validator=None):
+
+        if not object_validator:
+            if not self.config or not self.config.object_validator:
+                raise act.api.base.ArgumentError(
+                    "object_validator must be specified as argument or in configuration"
+                )
+
+            object_validator = self.config.object_validator
+
+        if self.source_object == self.destination_object:
+            raise act.api.base.ValidationError(
+                f"Source object can not be equal to destination object: {self.json()}"
+            )
+
+        if self.source_object and not object_validator(
+            self.source_object.type.name, self.source_object.value
+        ):
+            raise act.api.base.ValidationError(
+                f"Source object does not validate: {self.json()}"
+            )
+
+        if self.destination_object and not object_validator(
+            self.destination_object.type.name, self.destination_object.value
+        ):
+            raise act.api.base.ValidationError(
+                f"Destination object does not validate: {self.json()}"
+            )
+
+        return True
+
     def bidirectional(
         self,
         source_object_type,
@@ -535,6 +587,17 @@ class Fact(AbstractFact):
         params["destinationObject"] = object_serializer(self.destination_object)
         params["origin"] = origin_serializer(self.origin)
 
+        # Replace "" in value parameter with None
+        # This is for backward compatibilty for output produced on act-api <=v2.0.3
+        # Can be removed when we bump major version to 3.x.x
+        if params.get("value") == "":
+            warning(
+                "Replaced empty string in fact value with None. "
+                "This should be fixed in workers producing these facts, %s",
+                params,
+            )
+            params["value"] = None
+
         fact = self.api_post("v1/fact", **params)["data"]
 
         # Empty data and load new result from response
@@ -693,7 +756,7 @@ def fact_chain_seed(*facts):
     return "\n".join(sorted([str(fact) for fact in facts]))
 
 
-def fact_chain(*facts):
+def fact_chain(*facts: Fact) -> Iterable[Fact]:
     """
         Return fact chain of all facts.
         A fact chain is a modified version of the facts, where all placeholder
@@ -709,7 +772,7 @@ def fact_chain(*facts):
     # Calculate the hash for this chain, based on the seed
     sha256sum = hashlib.sha256(fact_chain_seed(*facts).encode("utf8")).hexdigest()
 
-    chain = copy.copy(facts)
+    chain: Iterable[Fact] = copy.copy(facts)
 
     # Return the fact chain, with all object values replaced "[placeholder[<HASH>]]"
     for fact in chain:
